/*

This code runs the arduino nano ble sens 33 that is used in the honey-chan product.


Version : 1_1_2
Created the : 22/11/2022
At : 4 place Jussieu
By Zeebee.

*/


/* --- Defines and Includes --- */

#define MAXIMWIRE_EXTERNAL_PULLUP
#include <Wire.h>
#include <MaximWire.h>
#include <DHT.h>
#include <DFRobot_INA219.h>
#include "HX711.h"

#define PIN_BUS_MAXIMWIRE     12
#define PIN_DHT_1             10
#define PIN_DHT_2             9
#define TYPE_DHT              DHT22
#define PIN_LOADCELL_DOUT     8 
#define PIN_LOADCELL_SCK      2
#define PIN_MICRO             A0
#define PIN_BATTERY           A1
#define PIN_INA219_SDA        A4
#define PIN_INA219_SCL        A5


#define SAMPLES 64
#define FREQ 1250
#define ANALOG_PIN A6

/* --- Variables --- */
//unsigned int T_envoie = 1000*60*10;
unsigned int T_envoie = 20000;

MaximWire::Bus                Bus_MaximWire(PIN_BUS_MAXIMWIRE);
MaximWire::DS18B20            maximwire_thermometre;
DHT                           dht22_1(PIN_DHT_1, TYPE_DHT);
DHT                           dht22_2(PIN_DHT_2, TYPE_DHT);
HX711                         loadcell;
DFRobot_INA219_IIC            ina219(&Wire, INA219_I2C_ADDRESS4);

float ina219Reading_mA = 99;
float extMeterReading_mA = 99;

int32_t Tab_Maxim_temp[2] = {9999, 9999}; 
int32_t dht22_1_temp = 9999; 
unsigned short dht22_1_hum = 9999;
int32_t dht22_2_temp = 9999;
unsigned short dht22_2_hum = 9999;
unsigned short weight = 9999;
unsigned short u_battery = 400;
unsigned short sunlight = 9999;


int32_t weight_offset = 0;
int32_t Tab_temp_int_1_2_offset[2] = {0, 0};
int32_t temp_int_3_offset = 0;
int32_t temp_ext_offset = 0;
int32_t hum_int_offset = 0;
int32_t hum_ext_offset = 0;

unsigned char flag_connected;



static char recv_buf[512];
static bool is_exist = false;
static bool is_join = false;
static int led = 0;
unsigned int ret = 0;
unsigned char led_flag = 0;


/* --- Fonctions --- */

// Fonction called in order to connect to the LoRaWan network
static int at_send_check_response(char *p_ack, int timeout_ms, char *p_cmd, ...)
{
    unsigned int ch;
    unsigned int num = 0;
    unsigned int index = 0;
    unsigned int startMillis = 0;

    memset(recv_buf, 0, sizeof(recv_buf));
    Serial1.write(p_cmd); 
    delay(200);
    startMillis = millis();

    do
    {
        while (Serial1.available() > 0)
        {
            ch = Serial1.read();
            recv_buf[index++] = ch;
            delay(2);
        }

    } while (millis() - startMillis < timeout_ms);

    if (strstr(recv_buf, p_ack) != NULL)
    {
        return 1;
    }
    else
        return 0;
}

// Fonction called in order to get the temperature of the maximwire temperature sensor
void Get_MaximWire_Temp()
{
  MaximWire::Discovery discovery = Bus_MaximWire.Discover();
  unsigned int cpt = 0;
    do
    {
        MaximWire::Address address;
        if (discovery.FindNextDevice(address))
        {
            if (address.GetModelCode() == MaximWire::DS18B20::MODEL_CODE)
            {
              MaximWire::DS18B20 maximwire_thermometre(address);

              float temp = maximwire_thermometre.GetTemperature<float>(Bus_MaximWire);
              Tab_Maxim_temp[cpt] = (temp + Tab_temp_int_1_2_offset[cpt] )* 10;

              cpt = cpt + 1;
              maximwire_thermometre.Update(Bus_MaximWire);
              delay(1000);
            }
        }
    } while (discovery.HaveMore());
}

// Fonction called in order to get the temperature and the humidity of the dht
void Get_DHT()
{
  float tauxHumidite = dht22_1.readHumidity();
  float temperatureEnCelsius = dht22_1.readTemperature();

  dht22_1_temp = ( temperatureEnCelsius + temp_int_3_offset )* 10;
  dht22_1_hum = ( tauxHumidite + hum_int_offset )* 10;

  tauxHumidite = dht22_2.readHumidity();
  temperatureEnCelsius = dht22_2.readTemperature();

  dht22_2_temp = ( temperatureEnCelsius + temp_ext_offset ) * 10;
  dht22_2_hum = ( tauxHumidite + hum_ext_offset ) * 10;
}

// Fonction called in order to get the weight sensored
void Get_Weight()
{
  if (loadcell.is_ready())
  {
      float weight_temp = (3.58e-5 * loadcell.read()) - 5.1225;
      weight = ( weight_temp + weight_offset ) * 100;
  }
}

// Fonction called in order to get the batterie voltage
void Get_U_Batterie()
{
  u_battery = (float) analogRead(PIN_BATTERY) / 2 * 4.2 / 10;
}

// Fonction called in order to get the sunlight power in lux
void Get_Sunlight()
{ 
  sunlight = 5*pow(10,(ina219.getBusVoltage_V()-0.4));
}

/* --- FFT relatives --- */
int data[SAMPLES];
unsigned short out[10];


float FFT(int in[], int N, float Frequency) {
  unsigned int Pow2[13] = {1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048};

    int a, c1, f, o, x;
    byte check = 0;
    a = N;

    for (int i = 0; i < 12; i++) {
        if (Pow2[i] <= a) {
            o = i;
        }
    }

    int out_r[Pow2[o]] = {};
    int out_im[Pow2[o]] = {};

    x = 0;
    for (int b = 0; b < o; b++) {
        c1 = Pow2[b];
        f = Pow2[o] / (c1 + c1);
        for (int j = 0; j < c1; j++) {
            x = x + 1;
            out_im[x] = out_im[j] + f;
        }
    }

    for (int i = 0; i < Pow2[o]; i++) {
        out_r[i] = in[out_im[i]];
        out_im[i] = 0;
    }

    int i10, i11, n1, tr, ti;
    float e;
    int c, s;
    for (int i = 0; i < o; i++) {
        i10 = Pow2[i];
        i11 = Pow2[o] / Pow2[i + 1];
        e = 360 / Pow2[i + 1];
        e = 0 - e;
        n1 = 0;

        for (int j = 0; j < i10; j++) {
            c = e * j;
            while (c < 0) {
                c = c + 360;
            }
            while (c > 360) {
                c = c - 360;
            }

            n1 = j;

            for (int k = 0; k < i11; k++) {
                if (c == 0) {
                    tr = out_r[i10 + n1];
                    ti = out_im[i10 + n1];
                } else if (c == 90) {
                    tr = -out_im[i10 + n1];
                    ti = out_r[i10 + n1];
                } else if (c == 180) {
                    tr = -out_r[i10 + n1];
                    ti = -out_im[i10 + n1];
                } else if (c == 270) {
                    tr = out_im[i10 + n1];
                    ti = -out_r[i10 + n1];
                } else if (c == 360) {
                    tr = out_r[i10 + n1];
                    ti = out_im[i10 + n1];
                } else if (c > 0 && c < 90) {
                    tr = out_r[i10 + n1] - out_im[i10 + n1];
                    ti = out_im[i10 + n1] + out_r[i10 + n1];
                } else if (c > 90 && c < 180) {
                    tr = -out_r[i10 + n1] - out_im[i10 + n1];
                    ti = -out_im[i10 + n1] + out_r[i10 + n1];
                } else if (c > 180 && c < 270) {
                    tr = -out_r[i10 + n1] + out_im[i10 + n1];
                    ti = -out_im[i10 + n1] - out_r[i10 + n1];
                } else if (c > 270 && c < 360) {
                    tr = out_r[i10 + n1] + out_im[i10 + n1];
                    ti = out_im[i10 + n1] - out_r[i10 + n1];
                }

                out_r[n1 + i10] = out_r[n1] - tr;
                out_r[n1] = out_r[n1] + tr;
                if (out_r[n1] > 15000 || out_r[n1] < -15000) {
                    check = 1;
                }

                out_im[n1 + i10] = out_im[n1] - ti;
                out_im[n1] = out_im[n1] + ti;
                if (out_im[n1] > 15000 || out_im[n1] < -15000) {
                    check = 1;
                }

                n1 = n1 + i10 + i10;
            }
        }

        if (check == 1) {
            for (int i = 0; i < Pow2[o]; i++) {
                out_r[i] = out_r[i] / 100;
                out_im[i] = out_im[i] / 100;
            }
            check = 0;
        }
    }

    int fout, fm, fstp;
    float fstep;
    fstep = Frequency / N;
    fstp = fstep;
    fout = 0;
    fm = 0;

    for (int i = 1; i < Pow2[o - 1]; i++) {
        if ((out_r[i] >= 0) && (out_im[i] >= 0)) {
            out_r[i] = out_r[i] + out_im[i];
        } else if ((out_r[i] <= 0) && (out_im[i] <= 0)) {
            out_r[i] = -out_r[i] - out_im[i];
        } else if ((out_r[i] >= 0) && (out_im[i] <= 0)) {
            out_r[i] = out_r[i] - out_im[i];
        } else if ((out_r[i] <= 0) && (out_im[i] >= 0)) {
            out_r[i] = -out_r[i] + out_im[i];
        }

        out_im[i] = out_im[i - 1] + fstp;
        if (fout < out_r[i]) {
            fm = i;
            fout = out_r[i];
        }

        out_r[i] = pow(out_r[i], 2);
    }

    int j = 0;
    for (int i = 0; i < 10; i++) {
        int n_f = 0;
        out[i] = 0;
        while ((out_im[j] - 48 * (i + 1) < 98) && j < data[o - 1]) {
            if (out_im[j] >= 98) {
                out[i] += out_r[j];
                n_f++;
            }
            j++;
        }
        if (n_f) out[i] = (char)out[i] / n_f;
    }

    float fa, fb, fc;
    fa = out_r[fm - 1];
    fb = out_r[fm];
    fc = out_r[fm + 1];
    fstep = (fa * (fm - 1) + fb * fm + fc * (fm + 1)) / (fa + fb + fc);

    return (fstep * Frequency / N);
}

void acquire() {
    for (int i = 0; i < 64; i++) {
        data[i] = analogRead(A0);
        delayMicroseconds(800);
    }
}

// Fonction called to Setup the LoRaWan 
void Setup_LoRaWan()
{
  ret = at_send_check_response("+AT: OK", 100, "AT\r\n");
  while ( !ret ){
    ret = at_send_check_response("+AT: OK", 100, "AT\r\n");
  }

  is_exist = true;
  at_send_check_response("+ID: AppEui", 1000, "AT+ID\r\n");
  at_send_check_response("+MODE: LWOTAA", 1000, "AT+MODE=LWOTAA\r\n");
  at_send_check_response("+DR: EU868", 1000, "AT+DR=EU868\r\n");
  at_send_check_response("+CH: NUM", 1000, "AT+CH=NUM,0-2\r\n");
  at_send_check_response("+KEY: APPKEY", 1000,
                          "AT+KEY=APPKEY,\"437BB90974XXXXXXXXXXXXXXXXXXXX\"\r\n");
  at_send_check_response("+KEY: DEVEUI", 1000, "AT+ID=DEVEUI,\"2022202220222022\"\r\n");
  at_send_check_response("+KEY: APPEUI", 1000, "AT+ID=APPEUI,\"0000000000000000\"\r\n");
  at_send_check_response("+CLASS: C", 1000, "AT+CLASS=A\r\n");
  at_send_check_response("+PORT: 8", 1000, "AT+PORT=8\r\n");
  delay(200);
}

// Fonction called in order to connect to lorawan network
void Connect_LoRaWan()
{
  if (at_send_check_response("+JOIN: Network joined", 12000, "AT+JOIN\r\n"))
  {
    is_join = true;
  }
  else
  {
    at_send_check_response("+ID: AppEui", 1000, "AT+ID\r\n");
    delay(5000);
  }
}

// Fonction called in order to send data though the LoRaWan network
void Communication_LoRaWan()
{
  char cmd[512];
  sprintf(cmd, "AT+MSGHEX=%04X%08X%08X%08X%04X%08X%04X%04X%04X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X\r\n",
                                                                                                            u_battery, 
                                                                                                            Tab_Maxim_temp[0], 
                                                                                                            Tab_Maxim_temp[1], 
                                                                                                            dht22_1_temp, 
                                                                                                            dht22_1_hum, 
                                                                                                            dht22_2_temp, 
                                                                                                            dht22_2_hum, 
                                                                                                            sunlight, 
                                                                                                            weight, 
                                                                                                            out[0], 
                                                                                                            out[1], 
                                                                                                            out[2], 
                                                                                                            out[3], 
                                                                                                            out[4], 
                                                                                                            out[5],
                                                                                                            out[6],
                                                                                                            out[7],
                                                                                                            out[8],
                                                                                                            out[9]
                                                                                                            );

  Serial.println(cmd);
  Serial.println(dht22_1_temp, HEX);

  ret = at_send_check_response("Done", 5000, cmd);
  if (ret)
  {
    delay(5000);
    Downlink_treatment(recv_buf);
  }
  else
  {
      Serial.print("Send failed!\r\n\r\n");
  }

}

void downlink_decoder(String str_data)
{
  String test = String (str_data.charAt(0));
  test.concat(str_data.charAt(1));

  if (str_data.length() > 0)
  {

    switch ( test.toInt() )
    {
      case 0 :
        T_envoie = 1000 * 6 * 20;
        break;

      case 1 :
        T_envoie = 1000 * 20;
        break;

      case 2 :
        T_envoie = 1000 * 60 * 2;
        break;

      case 3:
        T_envoie = 1000 * 60 * 20;
        break;

      default :
        break;
    }
    test.remove(0,2);

    test.concat(str_data.charAt(2));
    test.concat(str_data.charAt(3));
    switch ( test.toInt() )
    {
      case 0 :
        digitalWrite(LED_BUILTIN, LOW);
        break;

      case 1 :
        digitalWrite(LED_BUILTIN, HIGH);
        break;

      case 2 :
        digitalWrite(LED_BUILTIN, LOW);
        break;

      default :
        digitalWrite(LED_BUILTIN, LOW);
        break;
    }
    test.remove(0,2);

    test.concat(str_data.charAt(4));
    test.concat(str_data.charAt(5));
    switch ( test.toInt() )
    {
      case 0 :
        weight_offset = 0;
        break;

      case 1 :
        weight_offset += 1;
        break;

      case 2 :
        weight_offset -= 1;
        break;

      case 3 :
        weight_offset += 10;
        break;

      case 4 :
        weight_offset -= 10;
        break;
      
      default :
        weight_offset = 0;
        break;
    }
    test.remove(0,2);


    test.concat(str_data.charAt(6));
    test.concat(str_data.charAt(7));
    switch ( test.toInt() )
    {
      case 0 :
        Tab_temp_int_1_2_offset[0] = 0;
        break;

      case 1 :
        Tab_temp_int_1_2_offset[0] += 1;
        break;

      case 2 :
        Tab_temp_int_1_2_offset[0] -= 1;
        break;

      case 3 :
        Tab_temp_int_1_2_offset[0] += 10;
        break;

      case 4 :
        Tab_temp_int_1_2_offset[0] -= 10;
        break;
      
      default :
        Tab_temp_int_1_2_offset[0] = 0;
        break;
    }
    test.remove(0,2);


    test.concat(str_data.charAt(8));
    test.concat(str_data.charAt(9));
    switch ( test.toInt() )
    {
      case 0 :
        Tab_temp_int_1_2_offset[1] = 0;
        break;

      case 1 :
        Tab_temp_int_1_2_offset[1] += 1;
        break;

      case 2 :
        Tab_temp_int_1_2_offset[1] -= 1;
        break;

      case 3 :
        Tab_temp_int_1_2_offset[1] += 10;
        break;

      case 4 :
        Tab_temp_int_1_2_offset[1] -= 10;
        break;
      
      default :
        Tab_temp_int_1_2_offset[1] = 0;
        break;
    }
    test.remove(0,2);


    test.concat(str_data.charAt(10));
    test.concat(str_data.charAt(11));
    switch ( test.toInt() )
    {
      case 0 :
        temp_int_3_offset = 0;
        break;

      case 1 :
        temp_int_3_offset += 1;
        break;

      case 2 :
        temp_int_3_offset -= 1;
        break;

      case 3 :
        temp_int_3_offset += 10;
        break;

      case 4 :
        temp_int_3_offset -= 10;
        break;
      
      default :
        temp_int_3_offset = 0;
        break;
    }
    test.remove(0,2);


    test.concat(str_data.charAt(12));
    test.concat(str_data.charAt(13));
    switch ( test.toInt() )
    {
      case 0 :
        temp_ext_offset = 0;
        break;

      case 1 :
        temp_ext_offset += 1;
        break;

      case 2 :
        temp_ext_offset -= 1;
        break;

      case 3 :
        temp_ext_offset += 10;
        break;

      case 4 :
        temp_ext_offset -= 10;
        break;
      
      default :
        temp_ext_offset = 0;
        break;
    }
    test.remove(0,2);

    test.concat(str_data.charAt(14));
    test.concat(str_data.charAt(15));
    switch ( test.toInt() )
    {
      case 0 :
        hum_int_offset = 0;
        break;

      case 1 :
        hum_int_offset += 1;
        break;

      case 2 :
        hum_int_offset -= 1;
        break;

      case 3 :
        hum_int_offset += 10;
        break;

      case 4 :
        hum_int_offset -= 10;
        break;
      
      default :
        hum_int_offset = 0;
        break;
    }
    test.remove(0,2);


    test.concat(str_data.charAt(16));
    test.concat(str_data.charAt(17));
    switch ( test.toInt() )
    {
      case 0 :
        hum_ext_offset = 0;
        break;

      case 1 :
        hum_ext_offset += 1;
        break;

      case 2 :
        hum_ext_offset -= 1;
        break;

      case 3 :
        hum_ext_offset += 10;
        break;

      case 4 :
        hum_ext_offset -= 10;
        break;
      
      default :
        hum_ext_offset = 0;
        break;
    }
    test.remove(0,2);


    test.concat(str_data.charAt(18));
    test.concat(str_data.charAt(19));
    switch ( test.toInt() )
    {
      case 0 :
        digitalWrite(LED_PWR, LOW);
        break;

      case 1 :
        digitalWrite(LED_PWR, HIGH);
        break;

      case 2 :
        digitalWrite(LED_PWR, LOW);
        break;

      default :
        digitalWrite(LED_PWR, LOW);
        break;
    }
    test.remove(0,2);
  }
}

// Fonction called in order to receive data though the LoRaWan network
static void Downlink_treatment(char *p_msg)
{

  if (p_msg == NULL){ return;}

  char *p_start = NULL;
  char *char_data = NULL;
  String str_data;

  p_start = strstr(p_msg, "\"");
  p_start = strtok(p_start, "\"");
  char_data = strtok(p_start, "\"");

  str_data = String(char_data); 
  Serial.println(str_data);

  downlink_decoder(str_data);
  Serial.println(T_envoie);
  Serial.println("DONE");
}


/* --- Arduino default fonctions --- */
void setup()
{
  digitalWrite(LED_PWR, HIGH);
  digitalWrite(LED_BUILTIN, HIGH);
  

  delay(2000);
  Serial.begin(9600);
  Serial1.begin(9600);
  dht22_1.begin();
  dht22_2.begin();
  loadcell.begin(PIN_LOADCELL_DOUT, PIN_LOADCELL_SCK);

  Setup_LoRaWan();

  unsigned char flag = 0;
  unsigned char cpt = 0;
  while( flag ) {

    if ( ina219.begin() != true && cpt < 10)
    {
      cpt ++;
      delay(2000);
    }
    else
    {
      flag = 1;
    }

  }
  ina219.linearCalibrate(ina219Reading_mA, extMeterReading_mA);
  delay(2000);

  while ( is_join == false )
  {
    Connect_LoRaWan();
  }

  digitalWrite(LED_PWR, LOW);
  digitalWrite(LED_BUILTIN, LOW);
}

//
void loop()
{
  loadcell.power_up();
  delay(2000);

  digitalWrite(LED_PWR, HIGH);
  
  Get_MaximWire_Temp();
  Get_DHT();
  Get_Weight();
  Get_U_Batterie();  
  Get_Sunlight();

  acquire();
  FFT(data, SAMPLES, FREQ);

  Communication_LoRaWan();
  digitalWrite(LED_PWR, LOW);

  loadcell.power_down();
  delay(T_envoie - 2000);
}
